#include <poll.h>      /* For poll()   */
#include <stdio.h>     /* For perror() */
#include <unistd.h>    /* For sleep()  */

#include <vector>

using std::vector;

#include "headtask.hh"
#include "protocols.hh"
#include "socket.hh"
#include "stdin.hh"

typedef vector <struct HeadTask> HeadTaskListType;
typedef vector <struct Socket>   SocketListType;

static void Run(const HeadTaskListType &Tasks)
{
	SocketListType Sockets;
	
	StdinType Stdin;
	
	HeadTaskListType::const_iterator i;
	for(i=Tasks.begin(); i!=Tasks.end(); ++i)
	{
		switch(i->Protocol)
		{
			case SockProtocolTCP:
			{
				struct Socket sock;
				sock.BecomeListenerTCP(i->ListenPort);
				Sockets.push_back(sock);
				break;
			}
			case SockProtocolUDP:
			{
				struct Socket sock;
				sock.BecomeListenerUDP(i->ListenPort);
				Sockets.push_back(sock);
				break;
			}
		}
	}
	
	for(;;)
	{
		vector<struct pollfd> polls;
		
		typedef SocketListType::const_iterator SockCit;
		typedef SocketListType::iterator       SockIt;
		
		/* Add all sockets to the poll struct */
		
		for(SockCit i=Sockets.begin(); i!=Sockets.end(); ++i)
		{
			struct pollfd tmp = { i->fd, POLLIN, 0 };
			
			if(i->SendQueue.GetSize() > 0)
			{
				/* We want a chance to speak to the socket */
				tmp.events |= POLLOUT;
			}
			
			polls.push_back(tmp);
		}
		
		struct pollfd stdinfd = { STDIN_FILENO, POLLIN, 0 };
		struct pollfd stdoutfd = { STDOUT_FILENO, 0, 0 };

		if(Stdin.SendQueue.GetSize() > 0)
		{
			/* We want a chance to speak to stdout */
			stdoutfd.events |= POLLOUT;
		}
		
		/* Add stdin and stdout to the poll struct */
		
		polls.push_back(stdinfd);
		polls.push_back(stdoutfd);
		
		if(poll(&polls[0], polls.size(), -1) < 0)
		{
			perror("poll");
			/* FIXME: And now what? */
			continue;
		}
		
		for(unsigned a=0; a<polls.size(); ++a)
		{
			struct pollfd &tmp = polls[a];
			
			if(tmp.fd == STDIN_FILENO)
			{
				if(tmp.revents & (POLLNVAL | POLLHUP))
				{
					/* Master has disconnected! */
					
					SockIt i;
					for(i=Sockets.begin(); i!=Sockets.end(); ++i)
						close(i->fd);
					
					return;
				}
				if(tmp.revents & POLLIN)
				{
					/* Master wants to say something! */
					
					/* TODO: Stdin must now read fd 0. */
				}
			}
			else
			{
				SockIt i;
				for(i=Sockets.begin(); i!=Sockets.end(); ++i)
					if(i->fd==tmp.fd)
						break;
				
				if(i != Sockets.end())
				{
					Socket &sock = *i;
					if(tmp.revents & (POLLNVAL | POLLHUP))
					{
						/* This socket has died! */
						
						/* TODO: Must report it to the master */
						
						close(i->fd);
						Sockets.erase(i);
						continue;
					}
					
					switch(sock.Type)
					{
						case Socket::SockTypeListening:
						{
							switch(sock.Protocol)
							{
								case SockProtocolTCP:
								{
									if(tmp.revents & POLLIN)
									{
										Socket newsock;
										
										/* `Fwiffo! Fwiffo darling! Would you
										 *  please answer the door? I think
										 *  someone's there.'
										 */
										newsock.AcceptTCP(sock);
										
										Sockets.push_back(newsock);
										/* Invalidates sock (pointer) */
									}
									break;
								}
								case SockProtocolUDP:
								{
									if(tmp.revents & POLLIN)
									{
										/* UDP are connectionless sockets.
										 * You can directly try reading.
										 */
										
										sock.ReceiveUDP();
									}
									break;
								}
							}
							break;
						}
						case Socket::SockTypeConnected:
						{
							switch(sock.Protocol)
							{
								case SockProtocolTCP:
								{
									if(tmp.revents & POLLIN)
									{
										sock.ReceiveTCP();
									}
									break;
								}
								case SockProtocolUDP:
								{
									/* Empty: UDP sockets are never connected. */
									/* This case should never hit. */
									break;
								}
							}
							break;
						}
					} /* switch(sock.type) */
				} /* if(i != sockets.end()) */
			} /* if(fd != stdin_fileno) */
		} /* foreach poll */
	} /* for(;;) */
} /* Run() */


int main(void)
{
	HeadTaskListType HeadTasks;

	HeadTasks.push_back(HeadTask(6667, SockProtocolTCP));
	HeadTasks.push_back(HeadTask(4665, SockProtocolUDP));
	
	Run(HeadTasks);
	
	return 0;
}
